@@ -81,14 +81,13 @@
 #'   \item{\code{update(query, update = '{"$set":{}}', upsert = FALSE, multiple = FALSE)}}{Replace or modify matching record(s) with value of the \code{update} argument.}
 #' }
 #' @references Jeroen Ooms (2014). The \code{jsonlite} Package: A Practical and Consistent Mapping Between JSON Data and \R{} Objects. \emph{arXiv:1403.2805}. \url{http://arxiv.org/abs/1403.2805}
-mongo <- function(collection = "test",  db = "test", url = "mongodb://localhost", verbose = TRUE){
+mongo <- function(collection = "test", db = "test", url = "mongodb://localhost", verbose = TRUE){
   client <- mongo_client_new(url)
 
   # workaround for missing 'mongoc_client_get_default_database'
-  if(is.null(db) || db == "test"){
-    path <- get_path_from_url(url)
-    if(!is.null(path))
-      db <- path
+  if(missing(db) || is.null(db)){
+    if(!is.null(url_db <- mongo_get_default_database(client)))
+      db <- url_db
   }
 
   col <- mongo_collection_new(client, collection, db)
@@ -108,7 +107,8 @@ mongo_object <- function(col, client, verbose){
 
     iterate <- function(query = '{}', fields = '{"_id":0}', sort = '{}', skip = 0, limit = 0) {
       cur <- mongo_collection_find(col, query = query, sort = sort, fields = fields, skip = skip, limit = limit)
-      mongo_iterator(cur)
+      # make sure 'col' does not go out of scope to prevent gc
+      mongo_iterator(cur, col)
     }
 
     export <- function(con = stdout(), bson = FALSE){
