@@ -50,7 +50,7 @@ _mongoc_cursor_array_new (const char *field_name)
 
    ENTRY;
 
-   arr = bson_malloc0 (sizeof *arr);
+   arr = (mongoc_cursor_array_t *)bson_malloc0 (sizeof *arr);
    arr->field_name = field_name;
 
    RETURN (arr);
@@ -64,7 +64,7 @@ _mongoc_cursor_array_destroy (mongoc_cursor_t *cursor)
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
    if (arr->has_synthetic_bson) {
       bson_destroy(&arr->bson);
@@ -86,7 +86,7 @@ _mongoc_cursor_array_prime (mongoc_cursor_t *cursor)
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
    if (!arr->has_array) {
       arr->has_array = true;
 
@@ -113,7 +113,7 @@ _mongoc_cursor_array_next (mongoc_cursor_t *cursor,
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
    *bson = NULL;
 
    if (!arr->has_array) {
@@ -143,7 +143,7 @@ _mongoc_cursor_array_clone (const mongoc_cursor_t *cursor)
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
    clone_ = _mongoc_cursor_clone (cursor);
    _mongoc_cursor_array_init (clone_, arr->field_name);
@@ -161,7 +161,7 @@ _mongoc_cursor_array_more (mongoc_cursor_t *cursor)
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
    if (arr->has_array) {
       memcpy (&iter, &arr->iter, sizeof iter);
@@ -183,7 +183,7 @@ _mongoc_cursor_array_error  (mongoc_cursor_t *cursor,
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
    if (arr->has_synthetic_bson) {
       return false;
@@ -225,7 +225,7 @@ _mongoc_cursor_array_set_bson (mongoc_cursor_t *cursor,
 
    ENTRY;
 
-   arr = cursor->iface_data;
+   arr = (mongoc_cursor_array_t *)cursor->iface_data;
 
    bson_copy_to(bson, &arr->bson);
 
