@@ -43,6 +43,8 @@ typedef enum
    MONGOC_ERROR_COLLECTION,
    MONGOC_ERROR_GRIDFS,
    MONGOC_ERROR_SCRAM,
+   MONGOC_ERROR_SERVER_SELECTION,
+   MONGOC_ERROR_WRITE_CONCERN,
 } mongoc_error_domain_t;
 
 
@@ -92,6 +94,10 @@ typedef enum
    MONGOC_ERROR_QUERY_COMMAND_NOT_FOUND = 59,
    MONGOC_ERROR_QUERY_NOT_TAILABLE = 13051,
 
+   MONGOC_ERROR_SERVER_SELECTION_BAD_WIRE_VERSION,
+   MONGOC_ERROR_SERVER_SELECTION_FAILURE,
+   MONGOC_ERROR_SERVER_SELECTION_INVALID_ID,
+
    /* Dup with query failure. */
    MONGOC_ERROR_PROTOCOL_ERROR = 17,
 } mongoc_error_code_t;
